#![allow(dead_code)]

use crate::math::{Point, Vector, XformFull};

type FlagType = u8;
const OCCLUSION_FLAG: FlagType = 1;
const DONE_FLAG: FlagType = 1 << 1;

#[derive(Debug, Copy, Clone)]
pub struct Ray {
    pub orig: Point,
    pub dir: Vector,
    pub time: f32,
    pub wavelength: f32,
    pub max_t: f32,
    pub flags: FlagType,
}

/// A specifically local-space ray, for passing to functions when we've
/// already calculated the local-space version of a ray for the object
/// in question.
///
/// Also includes `dir_inv`, which is generally useful to have as well.
#[derive(Debug, Copy, Clone)]
pub struct LocalRay {
    pub orig: Point,
    pub dir: Vector,
    pub dir_inv: Vector,
}

impl Ray {
    pub fn new(
        orig: Point,
        dir: Vector,
        time: f32,
        wavelength: f32,
        max_t: f32,
        is_occlusion: bool,
    ) -> Self {
        Self {
            orig: orig,
            dir: dir,
            time: time,
            wavelength: wavelength,
            max_t: max_t,
            flags: if is_occlusion { OCCLUSION_FLAG } else { 0 },
        }
    }

    /// Creates a local ray from the given transform.
    pub fn to_local_xform(&self, xform: &XformFull) -> LocalRay {
        let orig = self.orig.xform_inv(xform);
        let dir = self.dir.xform_inv(xform);

        LocalRay {
            orig: orig,
            dir: dir,
            dir_inv: dir.recip(),
        }
    }

    /// Creates a local ray with no transform applied.
    pub fn to_local(&self) -> LocalRay {
        LocalRay {
            orig: self.orig,
            dir: self.dir,
            dir_inv: self.dir.recip(),
        }
    }

    //---------------------------------------------------------
    // Flags.

    /// Returns whether this is an occlusion ray.
    #[inline(always)]
    pub fn is_occlusion(&self) -> bool {
        (self.flags & OCCLUSION_FLAG) != 0
    }

    /// Returns whether this ray has finished traversal.
    #[inline(always)]
    pub fn is_done(&self) -> bool {
        (self.flags & DONE_FLAG) != 0
    }

    /// Marks this as an occlusion ray.
    #[inline(always)]
    pub fn mark_occlusion(&mut self) {
        self.flags |= OCCLUSION_FLAG
    }

    /// Marks this as having finished traversal.
    #[inline(always)]
    pub fn mark_done(&mut self) {
        self.flags |= DONE_FLAG
    }
}
